package org.skduemu.service;

import jakarta.xml.bind.JAXBContext;
import jakarta.xml.bind.Marshaller;
import org.eclipse.paho.client.mqttv3.*;
import org.skduemu.DTO.MqttClientDTO;
import org.skduemu.DTO.MqttDeviceDTO;
import org.skduemu.entity.*;
import org.skduemu.repository.TelemetryRepository;
import org.skduemu.xml.entity.Info;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

import java.io.*;
import java.sql.Timestamp;
import java.util.*;
import java.util.stream.Collectors;


@Service
public class MqttService {

    @Autowired
    private TelemetryRepository telemetryRepository;

    @Autowired
    private ModbusService modbusService;

    @Value("${coils:3}")
    private int countOfCoils;

    @Value("${udpIp:192.168.1.28}")
    private String host;

    @Value("${udpPort:8052}")
    private int port;
    
    @Value("${deviceIp:10.0.96.50}")
    private String deviceHost;
    @Value("${devicePort:19001}")
    private int devicePort;

    private Map<String, MqttClient> clients = new HashMap<>();

    private final Logger log = LoggerFactory.getLogger(this.getClass().getName());


    private void saveContent(MqttDevice mqttDevice) {
        Timestamp timestamp = new Timestamp(new Date().getTime());
        Telemetry telemetry = new Telemetry(timestamp, mqttDevice.getClientId(), mqttDevice.getContent().toString());
        telemetryRepository.save(telemetry);
    }

    public ResponseEntity<?> registerClients(MqttClientDTO mqttClientDTO) {
        try {
            List<MqttDeviceDTO> devicesDTO = mqttClientDTO.getMqttDeviceDTOList();
            for (MqttDeviceDTO deviceDTO : devicesDTO) {
                try {
                    MqttDevice listener = deviceDTO.getListener();
                    MqttDevice publisher = deviceDTO.getPublisher();
                    clients.put(listener.getClientId(), new MqttClient(listener.getBroker(), listener.getClientId()));
                    clients.put(publisher.getClientId(), new MqttClient(publisher.getBroker(), publisher.getClientId()));
                    boolean result = true;
                    MqttClient clientListener = clients.get(listener.getClientId());
                    MqttClient clientPublisher = clients.get(publisher.getClientId());
                    Boolean publisherOut = subscribeToTopicForPublisher(publisher, clientPublisher);
                    Boolean listenerOut = subscribeToTopic(listener, clientListener, publisher);
                    result = result & (listenerOut & publisherOut);
                    if (!result) {
                        return new ResponseEntity<>("Cannot subscribe client on current topic", HttpStatus.NOT_FOUND);
                    }
                    //return new ResponseEntity<>("OK", HttpStatus.OK);
                } catch (Exception e) {
                    e.printStackTrace();
                }
            }
        }
        catch (Exception e){
            return new ResponseEntity<>("Error", HttpStatus.NOT_FOUND);
        }
        return new ResponseEntity<>("Ok", HttpStatus.OK);
    }

    public Boolean subscribeToTopicForPublisher(MqttDevice mqttDevice, MqttClient client) {
        try {
            MqttConnectOptions mqttOptions = new MqttConnectOptions();
            client.connect(mqttOptions);
            if (client.isConnected()) {
                client.subscribe(mqttDevice.getTopic(), mqttDevice.getQos());
                MqttMessage mqttMsg = new MqttMessage(mqttDevice.getContent().toString().getBytes());
                mqttMsg.setQos(mqttDevice.getQos());
                log.info("Publisher connected");
            }
            return true;

        } catch (MqttException e) {
            e.printStackTrace();
            log.info(e.getMessage());
            return false;
        }
    }

    public Boolean subscribeToTopic(MqttDevice mqttDevice, MqttClient client, MqttDevice publisher) {
        try {
            MqttConnectOptions mqttOptions = new MqttConnectOptions();
            client.connect(mqttOptions);
            if (client.isConnected()) {
                client.setCallback(new MqttCallback() {
                    @Override
                    public void messageArrived(String topic, MqttMessage message) {
                        String payload = new String(message.getPayload());
                        payload = payload.replaceAll("[\\[\\]]", "");
                        List<Short> values = Arrays.stream(payload.split(", ")).map(Short::valueOf).collect(Collectors.toList());
                        modbusService.mqttToModbus(values);
                        publish(publisher.getClientId());
                        //sendUDP(new String(message.getPayload()));
                        saveContent(mqttDevice);
                        log.info("Received message: " + new String(message.getPayload()));
                    }

                    @Override
                    public void connectionLost(Throwable cause) {
                        cause.printStackTrace();
                        log.info("Connection is lost: " + cause.getMessage());
                    }

                    @Override
                    public void deliveryComplete(IMqttDeliveryToken token) {
                        log.info("Message publish is complete: " + token.isComplete());
                    }
                });

                client.subscribe(mqttDevice.getTopic(), mqttDevice.getQos());
                MqttMessage mqttMsg = new MqttMessage(mqttDevice.getContent().toString().getBytes());
                mqttMsg.setQos(mqttDevice.getQos());
            }
            log.info("Listener connected");
            return true;

        } catch (MqttException e) {
            e.printStackTrace();
            log.info(e.getMessage());
            return false;
        }
    }

    public ResponseEntity<?> disconnectClient() {
        try {
            for(Map.Entry<String, MqttClient> clientEntry : clients.entrySet()){
                MqttClient client = clientEntry.getValue();
                if(client.isConnected()){
                    client.disconnect();
                    log.info(client.getClientId() + " " + "disconnected from broker");
                }
            };
            return new ResponseEntity<>("Successful", HttpStatus.OK);
        }
        catch (Exception e){
            e.printStackTrace();
            return new ResponseEntity<>("Error. Cannot delete MqttClient", HttpStatus.NOT_FOUND);
        }
    }

    public ResponseEntity<?> publish(String clientId){
        try {
            MqttClient client = clients.get(clientId);
            String coils =  modbusService.readCoilStatus(0, countOfCoils);
            client.publish(clientId, new MqttMessage(coils.getBytes()));
            sendUDP(coils);
            return new ResponseEntity<>("Message publish is complete", HttpStatus.OK);
        }
        catch (Exception e){
            e.printStackTrace();
            return new ResponseEntity<>("Error. Message publishing isn't complete", HttpStatus.NOT_FOUND);
        }
    }
 
    public void sendUDP(String message){
        try{
            Info info = new Info();
            JAXBContext context = JAXBContext.newInstance(Info.class);
            StringWriter writer = new StringWriter();
            Marshaller marshaller = context.createMarshaller();
            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
            info.setUuid("KSK-MSUEV-04b-003d");
            info.setUrl(message);
            info.setTtl("101");
            marshaller.marshal(info, writer);
            sendMessage(String.valueOf(writer));
        }
        catch (Exception e){
            e.printStackTrace();
        }
    }

    @Scheduled(fixedRate = 5000) 
    private void upvsDevicePinging() {
    	sendUDP(deviceHost + ":" + devicePort);
    }
    
    
    public void sendMessage(String message){
        Sender sender = new Sender(host, port);
        sender.sendMessage(message);
    }

    //    public void unmarshaller(Scpd custom) {
//        try {
////            BufferedReader br = new BufferedReader(new FileReader("src/main/resources/static/body.xml"));
////            String body = br.lines().collect(Collectors.joining());
////            StringReader reader = new StringReader(body);
////            JAXBContext context = JAXBContext.newInstance(Scpd.class);
////            Unmarshaller unmarshaller = context.createUnmarshaller();
////            Scpd scpd = (Scpd) unmarshaller.unmarshal(reader);
//            Scpd scpd = new Scpd();
//            log.info(String.valueOf(custom.getActionList()));
//            JAXBContext context = JAXBContext.newInstance(Scpd.class);
//            StringWriter writer = new StringWriter();
//            Marshaller marshaller = context.createMarshaller();
//            marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, Boolean.TRUE);
//            scpd.setScpd_urn("Идентификатор устройства");
//            Argument argument = new Argument();
//            Argument argument1 = new Argument();
//            argument.setName("Имя 1");
//            argument1.setName("Имя 2");
//            argument.setDirection("Направление 1");
//            argument1.setDirection("Направление 2");
//            List<Argument> argumentList = new ArrayList<>();
//            argumentList.add(argument);
//            argumentList.add(argument1);
//            Action action = new Action();
//            Action action1 = new Action();
//            action.setName("Действие 1");
//            action1.setName("Действие 2");
//            action.setArgumentList(argumentList);
//            action1.setArgumentList(argumentList);
//            List<Action> actionList = new ArrayList<>();
//            actionList.add(action);
//            actionList.add(action1);
//            scpd.setActionList(actionList);
//            scpd.getActionList().get(0).getArgumentList().get(0).setName("pogremushka");
//            marshaller.marshal(scpd, writer);
//            System.out.println(writer);
//        } catch (Exception e) {
//            e.printStackTrace();
//        }
//    }
}
